/* Copyright (c) 1999 Christopher Y. Wong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import java.text.*;
import java.io.*;
import java.util.*;

class BookmarkModel extends Observable implements Serializable
{
  public BookmarkModel(File file) 
  {
    this.file = file;
    bookmarks = new Vector();
    checkpoint = 0;
    scanned = false;
  }
  
  public void setDocReaderModel(DocReaderModel docReaderModel)
  { this.docReaderModel = docReaderModel; }

  public void addBookmark(String name)
  {
    bookmarks.addElement(new Bookmark(name, docReaderModel.getCurrentLoc()));
    setChanged();
    notifyObservers();
  }

  public void setRecordLengths(int recordLengths[])
  {
    this.recordLengths = recordLengths;
  }

  public int[] getRecordLengths() { return recordLengths; }

  public void deleteBookmark(int idx)
  {
    try {
      bookmarks.removeElementAt(idx);
      setChanged();
      notifyObservers();
    } catch (ArrayIndexOutOfBoundsException e) {}
  }

  public Bookmark getBookmark(int idx)
  {
    try {
      return (Bookmark) bookmarks.elementAt(idx);
    } catch (ArrayIndexOutOfBoundsException e) {
      return null;
    }
  }

  public void goCheckpoint() { docReaderModel.setCurrentLoc(checkpoint); }

  public void goBookmark(int idx)
  {
    Bookmark bookmark;

    try {
      bookmark = (Bookmark) bookmarks.elementAt(idx);
    } catch (ArrayIndexOutOfBoundsException e) {
      return;
    }
    docReaderModel.setCurrentLoc(bookmark.offset);
  }

  public void update() 
  { 
    setChanged();
    notifyObservers();
  }

  public boolean save()
  {
    try {
      checkpoint = docReaderModel.getCurrentLoc();
      FileOutputStream fos = new FileOutputStream(file);
      ObjectOutputStream ois = new ObjectOutputStream(fos);
      ois.writeObject(this);
      ois.flush();
      ois.close();
      return true;
    } catch (IOException e) {
      System.out.println("Oops: " + e.toString());
      return false;
    }
  }

  public boolean scanned() { return scanned; }

  public void scan(DocIterator docIter)
  {
     if (scanCreatedBookmarks(docIter))
	return;
     
    // search backwards at most 80 characters for end of <...>
    int charsSearched = 0;
    char ch = docIter.last();
    while (ch != '>' && Character.isWhitespace(ch) && charsSearched < 80 && 
	   docIter.getIndex() > docIter.getBeginIndex()) {
      ch = docIter.previous();
      charsSearched++;
    }
    if (ch != '>') return;

    // auto-bookmark end delimiter found. Get bookmark marker.
    charsSearched = 0;
    StringBuffer marker = new StringBuffer();
    ch = docIter.previous();
    while (ch != '<' && charsSearched < 80 && ch != CharacterIterator.DONE) {
      marker.insert(0, ch);
      ch = docIter.previous();
    }
    if (ch != '<' || ch == CharacterIterator.DONE) return;

    // bookmark marker found. From top of doc, find all occurrences of
    // marker and get bookmark from that spot.
    int i, found = 0;
    int markerLoc = docIter.getIndex()+1;
    String curText, markerStr = marker.toString();
    docReaderModel.goTop();
    while (docReaderModel.seek(markerStr, false) && 
	   docReaderModel.getCurrentLoc() < markerLoc && 
	   found < 50) {
      docReaderModel.setCurrentLoc(docIter.getIndex());
      curText = docReaderModel.getText();
      // find eol or 80 characters whichever is shorter.
      i = 0;
      while (i<80 && i<curText.length() && curText.charAt(i) != '\n')
	i++;
      addBookmark(curText.substring(markerStr.length(), i));
      docReaderModel.lineDown();
      found++;
    }
    docReaderModel.goTop();
    scanned = true;
  }

   private boolean scanCreatedBookmarks(DocIterator docIter)
   {
      StringBuffer label = new StringBuffer();
      byte buf[] = new byte[18];
      int recPos, bmkPos;
      int numRecs = docIter.getNumTextRecs();
      int numPRCRecs = docIter.getNumPRCRecs();
      boolean compressed = docIter.isCompressed();
      RandomAccessFile raf = docIter.getDocFile();

      if (numPRCRecs - numRecs < 2)
	 return false; // no embedded bmks at end

      try {
	 for (int i=numRecs; i<numPRCRecs-1; i++) {
	    label.setLength(0);

	    // find offset of bookmark record.
	    raf.seek(0x56 + 8*i);
	    recPos = raf.readInt();

	    // read bookmark label (16 characters, null padded)
	    raf.seek(recPos);
	    raf.readFully(buf, 0, 16);
	    for (int j=0; j<16 && buf[j]!=0; j++)
	       label.append((char) (buf[j] & 0xff));

	    // read bookmark location (4 byte int)
	    bmkPos = raf.readInt();
	    
	    bookmarks.addElement(new Bookmark(label.toString(), bmkPos));
	 }
      } catch (IOException e) {
	 return false;
      }
      scanned = true;
      return true;
   }
  
  private transient DocReaderModel docReaderModel;
  private File file;
  private Vector bookmarks;
  private int checkpoint;
  private boolean scanned;
  private int recordLengths[];
  static final long serialVersionUID = 1L;
}
