/* Copyright (c) 1999 Christopher Y. Wong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import java.text.*;
import java.io.*;

public class DocIterator implements CharacterIterator
{
  public DocIterator(File file) throws IOException
  {
    this.file = file; 
    raf = new RandomAccessFile(file, "r");
    PRCHeader prcHeader = new PRCHeader(raf);
    if (!prcHeader.isDocFile())
      throw new IOException("Not a Doc file");
    numPRCRecs = prcHeader.getNumRecs();
      
    int docHeaderLoc = raf.readInt();
    raf.seek(docHeaderLoc);
    DocHeader docHeader = new DocHeader(raf);
    
    title = prcHeader.getTitle();
    compressed = docHeader.isCompressed();
    // getStoryLen() is not accurate, due to CR/LF conversions
    // textLength = docHeader.getStoryLen(); 
    numRecs = docHeader.getNumRecs();
    recSize = docHeader.getRecSize();

    index = 0;
    textLength = 0;

    docRecord = new DocRecord(raf, recSize, numRecs, numPRCRecs, compressed);
    if (loadBookmarks()) {
      // successfully loaded bookmarks. Use record lengths saved from
      // previous session.
      recordLengths = bookmarks.getRecordLengths();
      for (int i=0; i<numRecs; i++)
	textLength += recordLengths[i];
    } else {
      // no bookmark file. Find record lengths.
      recordLengths = new int[numRecs];
      for (int i=0; i<numRecs; i++) {
	docRecord.read(i);
	recordLengths[i] = docRecord.textLength();
	textLength += recordLengths[i];
      }
      bookmarks.setRecordLengths(recordLengths);
    }
    
    // preload first and (if available) second records
    currentOffset = 0;
    docRecord.read(0);
    otherRecord = new DocRecord(raf, recSize, numRecs, numPRCRecs, compressed);
    otherRecord.read(numRecs>1 ? 1 : 0);
  }

  public String getTitle() { return title; }
  public int getNumRecords() { return numRecs; }

  public final char first()
  {
    setIndex(getBeginIndex());
    return current();
  }

  public final char last()
  {
    setIndex(getEndIndex()-1);
    return current();
  }

  public final char current()
  {
     if (index < getBeginIndex() || getEndIndex()-1 < index)
	return DONE;
     else 
	return text.charAt(currentOffset);
  }

  public final char next()
  {
    if (index+1 >= getEndIndex()) {
      index = getEndIndex();
      return DONE;
    } else {
      setIndex(index+1);
      return current();
    }
  }

  public final char previous()
  {
    if (index-1 < getBeginIndex()) {
      index = getBeginIndex();
      return DONE;
    } else {
      setIndex(index-1);
      return current();
    }
  }

  public final char setIndex(int position)
  {
    if (position < getBeginIndex() || getEndIndex()-1 < position) 
      throw new IllegalArgumentException();

    try {
      // figure out which record we should load. Often, the records are
      // of variable sizes, so we have to sum up their sizes.
      int i=0, subtotal=0;
      while (position > subtotal-1) {
	subtotal += recordLengths[i];
	i++;
      }
    
      goRecord(i-1);
      index = position;
      currentOffset = position - (subtotal - recordLengths[i-1]);
    } catch (IOException e) {
      System.out.println("Caught Exception: " + e);
      e.printStackTrace();
    }
    return current();
  }

  public final int getBeginIndex() { return 0;}

  public final int getEndIndex() { return (int) textLength; }

  public final int getIndex() { return index; }

  public final Object clone()
  {
    System.out.println("Cloned !");

     try {
	DocIterator di = new DocIterator(file);
	di.setIndex(getIndex());
	return di;
     } catch (IOException e) {
       System.out.println("Caught Exception: " + e);
       e.printStackTrace();
       return null;
    }
  }

  public void goRecord(int record) throws IOException
  {
    if (docRecord.recordNumber() != record) {
      if (otherRecord.recordNumber() == record) {
	// desired loc is in other record. Swap them.
	DocRecord temp = docRecord;
	docRecord = otherRecord;
	otherRecord = temp;
      } else {
	// record unavailable: load it and save previous rec.
	DocRecord temp = docRecord;
	docRecord = otherRecord;
	otherRecord = temp;
	docRecord.read(record);
	text = docRecord.getText();
      }
    }
    text = docRecord.getText();
  }

  public BookmarkModel getBookmarks() { return bookmarks; }

   // exposed for BookmarkModel
   public RandomAccessFile getDocFile() { return raf; }
   public int getNumTextRecs() { return numRecs; }
   public int getNumPRCRecs() { return numPRCRecs; }
   public boolean isCompressed() { return compressed; }
   
  private boolean loadBookmarks()
       // load bookmark file corresponding to current file. Return
       // true if successful, false otherwise (creating a blank
       // bookmark model).
  {
    String docPath = file.getAbsolutePath();
    StringBuffer bmkPath = new StringBuffer(docPath);
    if (docPath.endsWith(".prc") || docPath.endsWith(".pdb")) {
      // replace pdb/prc suffix with bmk suffix
      int endIdx = bmkPath.length();
      bmkPath.setCharAt(--endIdx, 'k');
      bmkPath.setCharAt(--endIdx, 'm');
      bmkPath.setCharAt(--endIdx, 'b');
    } else
      bmkPath.append(".bmk"); // unknown suffix: just append our suffix.
    
    // see if we can load previously created bookmarks
    try {
      FileInputStream fis = new FileInputStream(bmkPath.toString());
      ObjectInputStream ois = new ObjectInputStream(fis);
      bookmarks = (BookmarkModel) ois.readObject();
      ois.close();
      return true;
    } catch (Exception e) {
      // could not open. Create fresh BookmarkModel
      bookmarks = new BookmarkModel(new File(bmkPath.toString()));
      return false;
    }
  }
  
  private BookmarkModel bookmarks;
  private int index, currentOffset;
  private String title, text;
  private boolean compressed;
  private long textLength;
  private int numRecs, numPRCRecs, recSize, firstRecLoc;
  private File file; 
  private RandomAccessFile raf;
  private DocRecord docRecord, otherRecord;
  private int recordLengths[];
}
