/* Copyright (c) 1999 Christopher Y. Wong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import java.text.*;
import java.io.*;
import java.util.*;

public class DocReaderModel extends Observable
{
  public DocReaderModel(ViewBounds viewBounds)
  {
    this.viewBounds = viewBounds;
    isValid = false;
    message = "No file loaded";
    text = "";
    line = new StringBuffer(128); // to hold at least 1 line of text
    lines = new StringBuffer(8192); // to hold entire page
    prevBoundaries = new IntStack(100);
    prevBreaks = new IntStack(100);
  }

  public BookmarkModel getBookmarks() { return bookmarks; }

  public void open(String filepath)
  {
    try {
      scanning = false;
      if (bookmarks != null)
	bookmarks.save();
      docIter = new DocIterator(new File(filepath));
      isValid = true;
      lastSeekLoc = 0;
      bookmarks = docIter.getBookmarks();
      bookmarks.setDocReaderModel(this);
      bookmarks.goCheckpoint();
      loadPage();
    } catch (IOException e) {
      isValid = false;
      message = "File open error: " + e.toString();
      update();
      return;
    }
    // scan for built-in bookmarks if we have not done so before
    if (!bookmarks.scanned()) {
      scanning = true;
      bookmarks.scan(docIter);
      scanning = false;
    }
    message = docIter.getTitle();
    update();
  }

  public void setCurrentLoc(int offset)
  {
     if (isValid) {
	loc = offset;
	loadPage();
     }
  }
      
  public void goTop()
  {
    if (isValid) {
      loc = 0;
      loadPage();
    }
  }

  public void goBottom()
  {
    if (isValid) {
      loc = docIter.getEndIndex() - 1;
      pageUp();
      loadPage();
    }
  }
  
  public void lineDown()
  {
    if (isValid) {
      if (text.lastIndexOf('\n', text.length()-2) == -1)
	return; // only 1 line onscreen: do not go down further
      int lfIdx = text.indexOf('\n');

      // eliminate first line from buffer and append terminating LF
      lines.setLength(0);
      lines.append(text.substring(lfIdx+1));
      lines.append('\n');
      
      // append additional line
      docIter.setIndex(lastLineLoc);
      goNextLine(line); // last line of current screen
      int newLastLineLoc = docIter.getIndex();
      if (newLastLineLoc < docIter.getEndIndex()) {
	// append next line to screen.
	lastLineLoc = newLastLineLoc;
	goNextLine(line); 
	lines.append(line);
      }
      lfIdx = lines.length()-1;
      if (lines.charAt(lfIdx) == '\n')
	lines.setLength(lfIdx);
      text = lines.toString();

      // move current location pointer down one line
      docIter.setIndex(loc);
      goNextLine(line);
      if (docIter.getIndex() < docIter.getEndIndex())
	loc = docIter.getIndex();
      update();
    }
  }

  public void pageDown()
  {
    // do not pgDn if nothing loaded or less than 2 lines displayed
    if (isValid && text.lastIndexOf('\n', text.length()-2)!=-1) {
      loc = lastLineLoc;
      loadPage();
    }
  }

  // loads viewBounds.getRows() lines into main buffer, wrapping lines
  // as needed. Lines are LF terminated.
  public void loadPage()
  {
    int i, eol, rows, prevBOL;

    if (!isValid) return;

    prevBOL = loc;
    rows = viewBounds.getRows();
    docIter.setIndex(loc);
    lines.setLength(0);
    for (i = 0; i < rows; i++) {
      prevBOL = docIter.getIndex();
      goNextLine(line);
      lines.append(line);
      eol = lines.length()-1;
      
      // test if we have reached EOF
      if (lines.charAt(eol) == CharacterIterator.DONE) {
	lines.setCharAt(eol, '\n');
	break;
      }

      // make sure line ends with an LF.
      if (lines.charAt(eol) != '\n')
	lines.append('\n');
    }
    lines.setLength(lines.length()-1); // chop off last LF.
    text = lines.toString();
    lastLineLoc = prevBOL;
    if (lastLineLoc == docIter.getEndIndex())
      lastLineLoc--;

    update();
  }

   public void lineUp()
   {
     if (isValid && loc > 0) {
       docIter.setIndex(loc);
       goPreviousLine(line, false);
       loc = docIter.getIndex();
       loadPage();
     }
   }
   
   public void pageUp()
   {
      if (isValid && loc > 0) {
	 lastLineLoc = loc;
	 docIter.setIndex(loc);
	 goNextLine(line); // include 1st line of current page
	 lines.setLength(0);
	 prevBreaks.empty();
	 int rows = viewBounds.getRows();
	 for (int i=0; i<rows; i++) {
	   goPreviousLine(line, true);
	   lines.insert(0, line);
	 }
	 loc = docIter.getIndex();
	 lines.setLength(lines.length()-1); // chop off trailing LF
	 text = lines.toString();

	 // special case: if we are at top, resync with top of page.
	 if (loc == 0)
	   loadPage();

	 update();
      }
   }

  public boolean doSearch(String pattern, boolean caseInsensitive)
  {
    if (isValid && seek(pattern, caseInsensitive)) {
      lastSeekLoc = docIter.getIndex();
      goPreviousLine(line, false);
      loc = docIter.getIndex();
      loadPage();
      return true;
    } else 
      return false;
  }

  public boolean doSearchAgain(String pattern, boolean caseInsensitive)
  {
    int curLoc = loc;
    loc = lastSeekLoc + 1;
    if (!doSearch(pattern, caseInsensitive)) {
      loc = curLoc;
      return false;
    } else
      return true;
    /*
    // unlike doSearch(), we want to skip past the first find.
    if (isValid && seek(pattern, caseInsensitive)) {
      goNextLine(line);
      loc = docIter.getIndex();
      return doSearch(pattern, caseInsensitive);
    } else
      return false;
      */
  }
  
  // seeks for text in document matching PATTERN, starting from
  // current location. Sets docIter to the start of the found text if
  // found, and returns true. If seek unsuccessful, return false and
  // resets docIter to the current location.
  public boolean seek(String pattern, boolean caseInsensitive)
  {
    // Partial implementation of Boyer and Moore string matching
    // algorithm: delta-1 and delta-2 shifts, but no extended delta-1.

    if (caseInsensitive)
      pattern = pattern.toUpperCase();

    if (pattern.length() == 1) {
      // special case: search for a single character. Skip the
      // overhead of the Boyer & Moore algorithm.
      docIter.setIndex(loc);
      return seekOneChar(pattern, caseInsensitive);
    }
    
    // first, analyze pattern to precompute delta-2 shifts.
    int idx;
    int patternLen = pattern.length();
    int d2shifts[] = new int[patternLen];
    for (int i=patternLen-2; i>=0; i--) {
      idx = pattern.lastIndexOf(pattern.substring(i+1), i);
      if (idx >= 0)
	d2shifts[i] = (i+1) - idx;
      else
	d2shifts[i] = patternLen;
    }
    d2shifts[patternLen-1] = 0;

    int i;
    char ch=0;
    int seekLoc = loc;
    int d1shift, nshift = 0;
    for (;;) {
      for (i = patternLen-1; i >= 0; i--) {
	try {
	  ch = docIter.setIndex(seekLoc+i);
	} catch (IllegalArgumentException e) {
	  // overshot end of file
 	  docIter.setIndex(loc);
	  return false;
	}
	if (ch == CharacterIterator.DONE) {
	  // reached end without finding search pattern.
	  docIter.setIndex(loc);
	  return false;
	}

	if (caseInsensitive)
	  ch = Character.toUpperCase(ch);
	if (ch != pattern.charAt(i)) 
	  break; // pattern mismatch found. Stop comparing.
      }

      // see if we have a match.
      if (i < 0) {
	// whole pattern matches. Set location and return success.
	loc = seekLoc;
	docIter.setIndex(loc);
	return true;
      }

      // find delta-1 shift
      nshift = pattern.lastIndexOf(ch, i);
      if (nshift == -1)
	nshift = i + 1;
      else
	nshift = patternLen - nshift -1;

      // shift by larger of delta-1 and delta-2 shifts
      if (i < patternLen-1)
	nshift = nshift > d2shifts[i] ? nshift : d2shifts[i];
      seekLoc += nshift;
      try {
	docIter.setIndex(seekLoc);
      } catch (IllegalArgumentException e) {
	// overshot end of file
	docIter.setIndex(loc);
	return false;
      }
    }
  }

  private boolean seekOneChar(String pattern, boolean caseInsensitive)
  {
    char searchCh = pattern.charAt(0);
    char ch = docIter.current();
    while (ch != searchCh && ch != CharacterIterator.DONE) {
      ch = docIter.next();
      if (caseInsensitive)
	ch = Character.toUpperCase(ch);
    }
    if (ch == CharacterIterator.DONE) {
      docIter.setIndex(loc);
      return false;
    } else 
      return true;
  }

  // move docIter to beginning of previous line. Fill LINE with characters
  // up to the new line, terminated with LF. If useStoredBreaks=true,
  // check prevBreaks for stored line breaks.
  private void goPreviousLine(StringBuffer line, boolean useStoredBreaks)
  {
    if (useStoredBreaks && !prevBreaks.isEmpty()) {
      // use one of the stored line breaks
      int newLoc = prevBreaks.pop();
      docIter.setIndex(newLoc);
      goNextLine(line);
      if (line.length() > 0 && line.charAt(line.length()-1) != '\n')
	line.append('\n');
      docIter.setIndex(newLoc);
      return;
    }
    
     line.setLength(0);
     int curLoc = docIter.getIndex();

     // backtrack iterator at least to previous line. Do this by going back
     // to just after the previous LF.
     docIter.previous();
     char ch = docIter.previous();
     while (ch != '\n' && ch != CharacterIterator.DONE)
	ch = docIter.previous();
     if (ch == '\n')
	docIter.next();

     // from this location, go forward line by line until we reach curLoc. At 
     // that point, we would have found the line that just precedes the
     // current location.
     int prevLineLoc = docIter.getIndex();
     while (docIter.getIndex() < curLoc) {
       prevLineLoc = docIter.getIndex();
       if (useStoredBreaks)
	 prevBreaks.push(prevLineLoc);
       goNextLine(line);
     }
     if (line.length() > 0 && line.charAt(line.length()-1) != '\n')
	line.append('\n');
     docIter.setIndex(prevLineLoc);

     // pop off the latest line break, so we do not repeat it next time
     if (useStoredBreaks && !prevBreaks.isEmpty()) 
       prevBreaks.pop();
  }

  // move docIter to beginning of next line. Fill LINE with characters
  // up to the new line. 
  private void goNextLine(StringBuffer line)
  {
    int estCols = viewBounds.getEstimatedColumns();
    int i, bol, prevBoundary;
    boolean fits;

    line.setLength(0);
    bol = docIter.getIndex();
    prevBoundaries.empty();

    // first pass: for efficiency. We fill the line with the estimated
    // number of columns. We will adjust later.
    char ch = docIter.current();
    line.append(ch);
    for (i=0; i < estCols; i++) {
      if (ch == ' ' || ch == '\t') // can wrap on these word breaks
	prevBoundaries.push(docIter.getIndex());
      if (ch == '\n' || ch == CharacterIterator.DONE) {
	// LF or EOF instantly ends a line
	docIter.next(); // jump to beg. of next line
	break; 
      }
      ch = docIter.next();
      line.append(ch);
    }
      
    if (i == estCols) {
      // we did not encounter LF. Keep adding words until we exceed
      // line length.
      ch = 0;
      while (viewBounds.lineFits(line)) {
	while (ch != ' ' && ch != '\t') {
	  ch = docIter.next();
	  line.append(ch);
	  if (ch == ' ' || ch == '\t') // can wrap on these word breaks
	    prevBoundaries.push(docIter.getIndex());
	  if (ch == '\n' || ch == CharacterIterator.DONE) {
	    // LF or EOF instantly ends a line
	    docIter.next(); // move to beg. of next line
	    return;
	  }
	} 
	ch = 0;
      }
    }

    while (!viewBounds.lineFits(line)) {
      // line too long: we need to wrap line at a previous word boundary 
      if (prevBoundaries.isEmpty()) {
	// no previous word break. Trim it by a character
	docIter.previous();
	line.setLength(line.length()-1); // chop off last char
      } else {
	prevBoundary = prevBoundaries.pop();
	// break line at last word boundary.
	docIter.setIndex(prevBoundary+1);
	line.setLength(prevBoundary - bol);
      }
    }
  }

  private void update()
  {
    if (isValid) {
      message = docIter.getTitle() + " (" + 
	(int) (( ((float) loc)/docIter.getEndIndex() ) * 100) + "%)";
      if (scanning)
	message = "Scanning: " + message;
      setChanged();
      notifyObservers();
    }
  }

  public String getText() { return text; }
  public void setMessage(String message) { this.message = message; }
  public String getMessage() { return message; }
  public int getCurrentLoc() { return loc; }
  public int getEndLoc() { return docIter.getEndIndex(); }
  public boolean isValid() { return isValid; }

  private int loc, lastLineLoc, lastSeekLoc;
  private IntStack prevBoundaries;
  // used by pageUp() in conjunction w/ goPreviousLine()
  private IntStack prevBreaks; 
  private boolean isValid, scanning;
  private DocIterator docIter;
  private ViewBounds viewBounds;
  private String message, text;
  private StringBuffer line, lines;
  private BookmarkModel bookmarks;
}

// a low-overhead implementation of an integer stack.
class IntStack
{
  public IntStack(int maxSize)
  {
    stack = new int[maxSize];
    items = 0;
  }

  public final void push(int val)
  {
    stack[items++] = val;
  }

  public final int pop()
  {
    return stack[--items];
  }

  public final boolean isEmpty() 
  {
    return items == 0;
  }

  public final void empty()
  {
    items = 0;
  }
  
  private int stack[];
  private int items;
}
