/* Copyright (c) 1999 Christopher Y. Wong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import java.text.*;
import java.io.*;

class DocRecord
{
  public DocRecord(RandomAccessFile raf, int recLen, int numRecs,
		   int numDBRecs, boolean compressed) 
  {
    this.raf = raf;
    this.recLen = recLen;
    this.numRecs = numRecs; // number of text records
    this.numDBRecs = numDBRecs; // number of records in PRC file
    this.compressed = compressed;
    decompText = "";
    text = new char[8192]; // rec sometimes exceeds recLen in size
    buf = new byte[8192];
  }

  public void read(int recNum) throws IOException
  {
    this.recNum = recNum;

    // find offset of record. This is the first 4-byte int of resource hdr
    raf.seek(0x56 + 8*recNum);
    int recPos = raf.readInt();
    
    // find length of record. This is recPos - next record's position.
    if (recNum == numDBRecs-2) {
      // last record. We use difference between recPos and eof
      recLen = (int) raf.length() - recPos;
    } else {
      raf.seek(0x56 + 8*(recNum+1)); // next resource header
      int recPos2 = raf.readInt();
      recLen = recPos2 - recPos;
    }

    raf.seek(recPos);
    raf.readFully(buf, 0, (int) recLen);
    if (compressed)
      decompress();
    else {
      // no compression: just copy buffer contents to text buffer
      for (int i=0; i<recLen; i++)
	text[i] = (char) (buf[i] & 0xff);
      decompText = new String(text, 0, recLen);
    }
  }

  /** read length of record, but do not actually load the record */
  public void scan(int recNum) throws IOException
  {
    // find offset of record. This is the first 4-byte int of resource hdr
    raf.seek(0x56 + 8*recNum);
    int recPos = raf.readInt();
    
    // find length of record. This is recPos - next record's position.
    if (recNum == numRecs-1) {
      // last record. We use difference between recPos and eof
      recLen = (int) raf.length() - recPos;
    } else {
      raf.seek(0x56 + 8*(recNum+1)); // next resource header
      int recPos2 = raf.readInt();
      recLen = recPos2 - recPos;
    }
  }
  
  public String getText() { return decompText; }

  public int recordNumber() { return recNum; }
  
  public int textLength() { return decompText.length(); }

  void decompress()
  {
    char c;
    int i=0; // input buffer index
    int o=0; // output buffer index

    while (i < recLen) {
      c = (char) (buf[i++] & 0xff);

      if (0<c && c<9) {
	// 1 ... 8  means copy that many bytes. For accented & binary chars
	while (c-- > 0)
	  text[o++] = ((char) (buf[i++] & 0xff));
      } else if (c < 0x80) {
	// 0, 9 ... 0x7f represent themselves
	text[o++] = c;
      } else if (c >= 0xc0) {
	// 0xc0 ... 0xff represents "space + ascii char"
	text[o++] = (char) 0x20; // space
	text[o++] = (char) (c ^ 0x80);
      } else {
	// 0x80 ... 0xbf represent sequences
	int m, n;
	c <<= 8;
	c += (buf[i++] & 0xff);
	m = (c & 0x3fff) >> COUNT_BITS;
	n = c & ((1<<COUNT_BITS) - 1);
	n += 3;
	while (n-- > 0) {
	  text[o] = (o-m >= 0) ? text[o-m] : ' ';
	  o++;
	}
      }
    }
    decompText = new String(text, 0, o);
  }

  private byte buf[];
  private char text[];
  private String decompText;
  private int recLen, numDBRecs, numRecs, recNum;
  private boolean compressed;
  private RandomAccessFile raf;
  private static final int COUNT_BITS = 3;
}
