/* Copyright (c) 1999 Christopher Y. Wong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.util.*;

class TextWidget extends Component 
  implements MouseListener, MouseMotionListener, ComponentListener
{
  public TextWidget()
    {
      setForeground(Config.getTextAreaFG());
      setBackground(Config.getTextAreaBG());
      setFont(Config.getTextAreaFont());
      lines = new Vector();

      addMouseListener(this);
      addMouseMotionListener(this);
      addComponentListener(this);

      dragging = false;
    }

  public final void setText(String text)
    {
      int start=0, end=0, lastEnd=0;

      // break text into lines, using LF as the line separator
      lines.setSize(0);
      end = text.indexOf('\n', start);
      while (end != -1) {
	lines.addElement(text.substring(start, end));
	start = end+1;
	if (start >= text.length())
	  break;
	lastEnd = end;
	end = text.indexOf('\n', start);
      }
      // tack on final line w/o terminating LF
      if (end == -1 && lastEnd+1 < text.length()-1)
	 lines.addElement(text.substring(lastEnd+1)); 

      repaint();
    }
  
   /** = number of rows of text that this widget can display */
   public final int getRows()
   {
      int height = getSize().height/fontHeight;
      return height>0 ? height-1 : height;
   }

   public void setFont(Font font)
   {
      super.setFont(font);
      fontHeight = getFontMetrics(font).getHeight();
      this.fm = Toolkit.getDefaultToolkit().getFontMetrics(font);
      charWidths = fm.getWidths();
   }

   public void paint(Graphics g)
   {
      int y;
      String line;

      if (offScreenImage == null) {
	 Dimension offScreenSize = getSize();
	 offScreenImage = createImage(offScreenSize.width,
				      offScreenSize.height);  
	 offScreenGraphics = offScreenImage.getGraphics();
      }
      
      Dimension size = getSize();
      offScreenGraphics.setColor(getBackground());
      offScreenGraphics.fillRect(0, 0, size.width, size.height);

      // draw outline
      offScreenGraphics.setColor(Color.black);
      offScreenGraphics.drawRect(0, 0, size.width-1, size.height-1);
      
      offScreenGraphics.setFont(getFont());
      offScreenGraphics.setColor(getForeground());

      for (int i=0; i<lines.size(); i++) {
	y = i*fontHeight + fontHeight;
	line = (String) lines.elementAt(i);
	offScreenGraphics.drawString(line, LEFT_MARGIN, y);
      }
      g.drawImage(offScreenImage, 0, 0, this);
   }

   public void mousePressed(MouseEvent e) 
   {
      if (lines.size() == 0)
	 return; // no text available
      
      // register start position of selection
      anchorRow = mouseRow(e.getY());
      anchorCol = mouseColumn(e.getX(), anchorRow);
      lastRow = anchorRow;
      lastCol = anchorCol;

      // begin selection highlighting
      flipH(anchorRow, anchorCol, anchorCol);
      dragging = true;
   }

   public void mouseReleased(MouseEvent e) 
   {
      if (!dragging) return;
      
      // figure out where the highlight starts and ends. The mouse pointer
      // may have dragged to a location before the anchor point.
      int curRow = mouseRow(e.getY());
      int curCol = mouseColumn(e.getX(), curRow);
      int startRow, endRow, startCol, endCol;
      if (curRow > anchorRow ||
	  (curRow == anchorRow && curCol > anchorCol)) {
	 startRow = anchorRow; startCol = anchorCol;
	 endRow = curRow; endCol = curCol;
      } else {
	 startRow = curRow; startCol = curCol;
	 endRow = anchorRow; endCol = anchorCol;
      }

      // don't select if mouse has not been moved from anchor position
      if (curRow == anchorRow && curCol == anchorCol) {
	 dragging = false;
	 repaint();
	 requestFocus();
	 return;
      }

      // determine selected text
      StringBuffer selected = new StringBuffer();
      String line = (String) lines.elementAt(startRow);
      if (startRow == endRow) {
	 endCol = endCol >= line.length() ? line.length() : endCol+1;
	 selected.append(line.substring(startCol, endCol));
      } else {
	 selected.append(line.substring(startCol));
	 selected.append('\n');
	 for (int i=startRow+1; i<endRow; i++) {
	    selected.append((String) lines.elementAt(i));
	    selected.append('\n');
	 }
	 line = (String) lines.elementAt(endRow);
	 endCol = endCol >= line.length() ? line.length() : endCol+1;
	 selected.append(line.substring(0, endCol));
      }

      // copy selected text to clipboard
      String str = selected.toString();
      StringSelection clipString = new StringSelection(str);
      this.getToolkit().getSystemClipboard().setContents(clipString, null);

      // clear selection highlight
      dragging = false;
      repaint();
      requestFocus();
   }

   public void mouseDragged(MouseEvent e) 
   {
      int col;
      String line;
      
      if (lines.size() == 0) return;
       
      // update selection highlight.
      dragging = true;
      int curRow = mouseRow(e.getY());
      int curCol = mouseColumn(e.getX(), curRow);

      // figure out where the highlight starts and ends. The mouse pointer
      // may have dragged to a location before the anchor point.
      int startRow, endRow, startCol, endCol;
      boolean forward; // direction of drag: forwards or backwards from anchor
      if (curRow > anchorRow ||
	  (curRow == anchorRow && curCol > anchorCol)) {
	 startRow = anchorRow; startCol = anchorCol;
	 endRow = curRow; endCol = curCol;
	 forward = true;
      } else {
	 startRow = curRow; startCol = curCol;
	 endRow = anchorRow; endCol = anchorCol;
	 forward = false;
      }

      if (curRow == lastRow && curCol == lastCol) return;

      // see if we need to clear highlight of previous position
      if (lastRow > endRow) {
	 flipH(lastRow, 0, lastCol);
	 for (int i=lastRow-1; i>=endRow; i--) {
	    line = (String) lines.elementAt(i);
	    if (i == anchorRow)
	       flipH(i, anchorCol, line.length()-1);
	    else
	       flipH(i, 0, line.length()-1);
	 }
      } else if (lastRow < startRow) {
	 line = (String) lines.elementAt(lastRow);
	 flipH(lastRow, lastCol, line.length()-1);
	 for (int i=lastRow+1; i<=startRow; i++) {
	    line = (String) lines.elementAt(i);
	    if (i == anchorRow) 
	       flipH(i, 0, anchorCol);
	    else
	       flipH(i, 0, line.length()-1);
	 }
      } else if (lastRow == startRow && lastCol < startCol) {
	 flipH(lastRow, lastCol, startCol-1);
      } else if (lastRow == endRow && lastCol > endCol) {
	 flipH(lastRow, endCol+1, lastCol);
      }

      // highlight rows up to but not including current row
      if (forward && endRow > lastRow && endRow > anchorRow) {
	 int beginHilitRow;
	 if (lastRow < anchorRow) {
	    beginHilitRow = anchorRow;
	    col = anchorCol;
	 } else {
	    beginHilitRow = lastRow;
	    col = lastRow==anchorRow && lastCol<anchorCol ?
	       anchorCol : lastCol+1;
	 }
	 line = (String) lines.elementAt(beginHilitRow);
	 flipH(beginHilitRow, col, line.length()-1);
	 for (int i=beginHilitRow+1; i<endRow; i++) {
	    line = (String) lines.elementAt(i);
	    flipH(i, 0, line.length()-1);
	 }
      } else if (!forward && startRow < lastRow && startRow != anchorRow) {
	 int bottomRow = lastRow>anchorRow ? anchorRow : lastRow;
	 col = bottomRow==anchorRow && anchorCol<lastCol ?
	    anchorCol : lastCol-1;
	 if (bottomRow==anchorRow && lastRow > anchorRow)
	    col = anchorCol;
	 flipH(bottomRow, 0, col);
	 for (int i=bottomRow-1; i>startRow; i--) { 
	    line = (String) lines.elementAt(i);
	    col = i == anchorRow ? anchorCol : line.length()-1;
	    flipH(i, 0, col);
	 }
      }
      
      // highlight text on current row up to current col.
      if (forward) {
	 if (curRow == lastRow)
	    col = lastCol+1;
	 else
	    col = curRow==anchorRow ? anchorCol : 0;
	 flipH(curRow, col, curCol);
      } else if (curRow == lastRow && curCol < lastCol) {
	 flipH(curRow, curCol, lastCol-1);
      } else if (curRow != lastRow && curRow == startRow) {
	 line = (String) lines.elementAt(startRow);
	 col = curRow==anchorRow ? anchorCol : line.length()-1;
	 flipH(curRow, curCol, col);
      }

      lastRow = curRow;
      lastCol = curCol;
   }

   private void flipH(int row, int startCol, int endCol)
   {
      Graphics g = getGraphics();
      g.setXORMode(getForeground());
      g.setColor(getBackground());

      int y = row * fontHeight + 5;
      int startX = screenColumn(row, startCol);
      int endX = screenColumn(row, endCol+1);
      g.fillRect(startX, y, endX-startX, fontHeight);
   }

   public void mouseEntered(MouseEvent e)
   { setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR)); }

   public void mouseExited(MouseEvent e)
   { setCursor(Cursor.getDefaultCursor()); }
   
   // = text row on screen indicated by mouse pointer position
   private int mouseRow(int y)
   {
      int row = y / fontHeight;
      return row >= lines.size() ? lines.size()-1 : (row>=0 ? row : 0);
   }
   
   // = text column on screen indicated by mouse pointer position
   private int mouseColumn(int x, int row)
   {
      if (x < LEFT_MARGIN) return 0; // before first column
      
      String line = (String) lines.elementAt(row);
      int cumWidth = 0;
      int i = 0;
      int lineLength = line.length();
      while (cumWidth < x-LEFT_MARGIN && i < lineLength) 
	 cumWidth += charWidths[line.charAt(i++)];
      if (i >= lineLength && i > 0)
	 return lineLength-1; // clicked past end of line.
      if (cumWidth >= x-LEFT_MARGIN)
	 return i>0 ? i-1 : 0;

      return 0;
   }

   private int screenColumn(int row, int col)
   {
      String line = (String) lines.elementAt(row);
      if (line.length() < col)
	 col = line.length();
      int cumWidth = 0;
      for (int i=0; i<col; i++)
	 cumWidth += charWidths[line.charAt(i)];
      return cumWidth + LEFT_MARGIN;
   }
   
   public void componentResized(ComponentEvent e)
   {
      offScreenImage    = createImage(getSize().width, getSize().height);
      offScreenGraphics = offScreenImage.getGraphics();
   }

   public boolean isFocusTraversable() { return true; }

   public void mouseClicked(MouseEvent e) {}
   public void mouseMoved(MouseEvent e) {}
   public void componentMoved(ComponentEvent e) {}
   public void componentShown(ComponentEvent e) {}
   public void componentHidden(ComponentEvent e) {}
   
   private Vector lines; // a vector of strings, one for each line.
   private int fontHeight; // of current font
   private int charWidths[];
   private FontMetrics fm;
   
   private int anchorRow, anchorCol, lastRow, lastCol;
   private boolean dragging;
   
   private Image offScreenImage;
   private Graphics offScreenGraphics;

   private final int LEFT_MARGIN = 5;
}

